#!/usr/bin/env dart

/**
 * Command line interface.
 * @module bin.cli
 */
library akismet.cli;

// Module dependencies.
import 'dart:io';
import 'package:akismet/io.dart' as akismet;
import 'package:args/args.dart';
import 'package:path/path.dart' as path;

/**
 * Represents an application providing functionalities specific to console requests.
 * @class cli.Application
 * @static
 */

/**
 * Object used to parse command line argments.
 * @property _parser
 * @type args.ArgParser
 * @static
 * @private
 */
final ArgParser _parser=new ArgParser()
  ..addFlag('help', abbr: 'h', help: 'output usage information', negatable: false)
  ..addFlag('version', abbr: 'V', help: 'output the version number', negatable: false)
  ..addOption('address', abbr: 'a', defaultsTo: '0.0.0.0', help: 'address that the server should run on')
  ..addOption('port', abbr: 'p', defaultsTo: '3000', help: 'port that the server should run on')
  ..addOption('redirect', abbr: 'r', help: 'the URL to redirect when a request is unhandled')
  ..addFlag('silent', help: 'silence the log output from the server', negatable: false);

/**
 * Value indicating whether to silence the log output from the server.
 * @property _silent
 * @type bool
 * @static
 * @private
 */
bool _silent=false;

/**
 * Prints the specified message, with a timestamp and a new line, to the standard output.
 * @method log
 * @param {String} message The message to be logged.
 * @static
 */
void log(String message) {
  if(!_silent) {
    var now=new DateTime.now();
    print('[$now] $message');
  }
}

/**
 * Application entry point.
 * @method main
 * @param {List&lt;String&gt;} arguments The command line arguments.
 * @static
 */
void main(List<String> arguments) {
  try {
    var results=_parser.parse(arguments);
    if(results['help']) return printUsage();
    if(results['version']) return print(akismet.VERSION);
    if(results['silent']) _silent=true;

    var redirectUrl=(results['redirect']!=null ? Uri.parse(results['redirect']) : null);
    return startServer(results['address'], int.parse(results['port']), redirectUrl: redirectUrl);
  }

  on FormatException catch(e) {
    print(e.message);
    exit(1);
  }
}

/**
 * Prints the usage information.
 * @method printUsage
 * @static
 */
void printUsage() {
  var script=path.basename(Platform.script.toFilePath());
  var buffer=new StringBuffer()
    ..writeln('Starts an Akismet.dart server.')
    ..writeln()
    ..writeln('Usage:')
    ..writeln('    $script [options]')
    ..writeln()
    ..writeln('Options:')
    ..write(_parser.getUsage());

  print(buffer);
}

/**
 * Starts a server listening for HTTP requests.
 * @method startServer
 * @param {String|InternetAddress} address The address that the server should run.
 * @param {int} port The port that the server is should run.
 * @param {Uri} [redirectUrl] The URL to redirect the user when a request is unhandled.
 * @static
 */
void startServer(address, int port, { Uri redirectUrl }) {
  var server=new akismet.Server(redirectUrl);

  server.onError.listen((error) {
    log('ERROR - $error');
  });

  server.onRequest.listen((HttpRequest request) {
    var address=request.connectionInfo.remoteAddress.address;
    var userAgent=request.headers[HttpHeaders.USER_AGENT].first;
    log('$address - "${request.method} ${request.uri} HTTP/${request.protocolVersion}" "$userAgent"');
  });

  server.bind(address, port).then((_) {
    log('Akismet server listening on $address:$port');
  });
}
