#!/usr/bin/env node
/* global cd, config, cp, echo, env, exec, exit, rm, target */

/**
 * Build system.
 * @module bin.make
 */
'use strict';

// Module dependencies.
require('shelljs/make');
var path=require('path');
var util=require('util');

/**
 * Provides tasks for [ShellJS](http://shelljs.org) make tool.
 * @class cli.Makefile
 * @static
 */
cd(__dirname+'/..');

/**
 * The application settings.
 * @property config
 * @type Object
 */
config.fatal=true;

/**
 * Runs the default tasks.
 * @method all
 */
target.all=function() {
  target.js();
};

/**
 * Builds the documentation.
 * @method doc
 */
target.doc=function() {
  echo('Build the documentation...');
  exec('docgen');
  cp('-f', [ 'www/apple-touch-icon.png', 'www/favicon.ico' ], 'doc/api/assets');
};

/**
 * Builds the client scripts.
 * @method js
 */
target.js=function() {
  echo('Build the client scripts...');
  if(!('DART_SDK' in env)) {
    echo('Unable to locate "dart2js" program: DART_SDK environment variable not set.');
    exit(1);
  }

  exec(util.format(
    '"%s" --minify --out="%s" "%s"',
    path.join(env.DART_SDK, 'bin/dart2js'),
    path.resolve('www/js/tests.js'),
    path.resolve('test/html_test.dart')
  ));
  
  rm('-f', 'www/js/tests.js.*');
  rm('-f', 'www/js/tests.precompiled.js');
};

/**
 * Performs static analysis of source code.
 * @method lint
 */
target.lint=function() {
  config.fatal=false;

  echo('Static analysis of JavaScript sources...');
  exec('jshint --exclude bin/packages --verbose bin');

  echo('Static analysis of documentation comments...');
  exec('docgen --lint');

  config.fatal=true;
};

/**
 * Runs the unit tests.
 * @method test
 */
target.test=function() {
  var program=require('commander');
  program._name='make test';
  
  program
    .version(require('../package.json').version)
    .option('-k, --key <apiKey>', 'the Akismet API key')
    .option('-b, --blog <url>', 'the front page or home URL [http://dev.belin.io/akismet.dart]', 'http://dev.belin.io/akismet.dart')
    .parse(process.argv);

  if(!program.key) program.help();
  
  echo('Run the unit tests...');
  if(!('DART_SDK' in env)) {
    echo('Unable to locate "dart" program: DART_SDK environment variable not set.');
    exit(1);
  }

  var result=exec(util.format(
    '"%s" "%s" %s %s',
    path.join(env.DART_SDK, 'bin/dart'),
    path.resolve('test/io_test.dart'),
    program.key,
    program.blog
  ));
  
  exit(result.code);
};
