/**
 * Implementation of the `akismet.Client` class.
 * @module core.client
 */
part of akismet.core;

/**
 * Provides the base class for clients that submit comments to [Akismet](https://akismet.com) service.
 * @class akismet.Client
 * @constructor
 * @param {String} apiKey The Akismet API key.
 * @param {Uri|akismet.Blog} blog The front page or home URL.
 */
abstract class Client {

  Client(this.apiKey, blog) {
    assert(apiKey!=null);
    assert(blog!=null);
    this.blog=(blog is Blog ? blog : new Blog(blog));
  }

  /**
   * The host of the default remote service.
   * @property DEFAULT_SERVICE
   * @type String
   * @static
   * @final
   */
  static const String DEFAULT_SERVICE='rest.akismet.com';

  /**
   * The Akismet API key.
   * @property apiKey
   * @type String
   */
  String apiKey;

  /**
   * The front page or home URL of the instance making requests.
   * @property blog
   * @type akismet.Blog
   */
  Blog blog;

  /**
   * The URL of the remote service.
   * @property serviceUrl
   * @type Uri
   * @default Uri.parse("https://rest.akismet.com")
   */
  Uri serviceUrl=Uri.parse('https://$DEFAULT_SERVICE');

  /**
   * The user agent string to use when making requests.
   * If possible, the user agent string should always have the following format: `Application Name/Version | Plugin Name/Version`.
   * @property userAgent
   * @type String
   */
  String userAgent;

  /**
   * Checks the specified comment against the service database.
   * @method checkComment
   * @param {akismet.Comment} comment The comment to be checked.
   * @return {Future&lt;bool&gt;} Value indicating whether it is spam.
   * @async
   */
  Future<bool> checkComment(Comment comment) {
    assert(comment!=null);

    var endPoint=(
      serviceUrl.host!=Client.DEFAULT_SERVICE ?
      serviceUrl.resolve(EndPoints.checkComment.pattern) :
      new Uri(
        scheme: serviceUrl.scheme,
        host:'$apiKey.${serviceUrl.host}',
        port: serviceUrl.port,
        path: EndPoints.checkComment.pattern
      )
    );

    return queryService(endPoint, comment.toJson()).then((result) => result=='true');
  }

  /**
   * Queries the remote service by posting the specified fields to a given end point.
   * @method queryService
   * @param {String} endPoint The URL of the end point to query.
   * @param {Map&lt;String, String&gt;} fields The fields describing the query body.
   * @return {Future&lt;String&gt;} The response as a string.
   * @async
   */
  Future<String> queryService(Uri endPoint, Map<String, String> fields);

  /**
   * Submits the specified comment that was incorrectly marked as spam but should not have been.
   * @method submitHam
   * @param {akismet.Comment} comment The comment to be submitted.
   * @return {Future} A `Future` that completes once the comment has been submitted.
   * @async
   */
  Future submitHam(Comment comment) {
    assert(comment!=null);

    var endPoint=(
      serviceUrl.host!=Client.DEFAULT_SERVICE ?
      serviceUrl.resolve(EndPoints.submitHam.pattern) :
      new Uri(
        scheme: serviceUrl.scheme,
        host:'$apiKey.${serviceUrl.host}',
        port: serviceUrl.port,
        path: EndPoints.submitHam.pattern
      )
    );

    return queryService(endPoint, comment.toJson());
  }

  /**
   * Submits the specified comment that was not marked as spam but should have been.
   * @method submitSpam
   * @param {akismet.Comment} comment The comment to be submitted.
   * @return {Future} A `Future` that completes once the comment has been submitted.
   * @async
   */
  Future submitSpam(Comment comment) {
    assert(comment!=null);

    var endPoint=(
      serviceUrl.host!=Client.DEFAULT_SERVICE ?
      serviceUrl.resolve(EndPoints.submitSpam.pattern) :
      new Uri(
        scheme: serviceUrl.scheme,
        host:'$apiKey.${serviceUrl.host}',
        port: serviceUrl.port,
        path: EndPoints.submitSpam.pattern
      )
    );

    return queryService(endPoint, comment.toJson());
  }

  /**
   * Returns a string representation of this object.
   * @method toString
   * @return {String} A string that represents the current object.
   */
  @override
  String toString() => 'Client { apiKey: "$apiKey", blog: "${blog.url}", serviceUrl: "$serviceUrl", userAgent: "$userAgent" }';

  /**
   * Checks the `apiKey` against the service database.
   * @method verifyKey
   * @return {Future&lt;bool&gt;} Value indicating whether it is a valid API key.
   * @async
   */
  Future<bool> verifyKey() {
    var endPoint=serviceUrl.resolve(EndPoints.verifyKey.pattern);
    return queryService(endPoint, { 'key': apiKey }).then((result) => result=='valid');
  }
}
